BBBBBBB BBBBBBBBBBBBBBBBBBBBBBBBBBB
BBBB BBBBBBBBBBBBBBBBB BBBB BBBBBBBBBBBBBBBBBBBB BBBBBB


BBBBB BBBBBBB

BB BBBB
    XXXXXXXXXXXXXXXX
    XXXXXX
        BBBBBB BBBB BBBBBBBBBBBBBBBBBBBBBB
        XXXXXX XXXXXXXXXXXXX XXXXXXXXXXXXXX XXXXXXXXXX XXXXXXXXXXX XXXXXX XXXXXXXXX XXXXXXXXXXXXXXX
    XXXXXXX
BBBBB
    XXX XXXXXXXXXXXXXXXXXXXXXXXXX
    XXXX XXXXXXXXXXXX XXXXXX
        XXXX XXXXXXXXXXXXXXXXXX
            XXXX XXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXX
                XXXX XXXXXXXXXXXXXXX XXXXXXXXXXXXX
                    XXX XXXXXXXXXXXXXXXXX XXXXXXXXXXXXX

                        BBB BBBBB BB BBBBBBBBBBBBBBBBBBBBBBBBBB
                            XXX XXXXXXXXXXXXXXXXXXXXXXXXX
                                XX XXXXXXXX XXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXX
                                    BB BBBBBBBBBB BB BBBBB
                                        XX XXXXXXXXX XXXXXXXXXXXXXXXXXX
                                    BBBB BBBBBBBBBB BB BBBBBB
                                        XX XXXXXXXXX XXXXXXXXXXXXXXXXXXX
                                    BBBB
                                        XX XXXXXXXXX XXXXXXXXXXXXXXXXXX
                                    BBBBB 
                                XXXXXXXXX
                        BBBBBB
                    XXXXX
                XXXXXX
                XXXX XXXXXXXXXXXXXXXXXX XXXXXXXXXXX XXXXXXXXXXXXXXXX
                BB BBBBBBBBBBBBBBBBBBBBBBBBBB
                    
                    BB BBBBBBBBBBBBBBBBBBBBBBBBB BB BBBBBBBBBB
                        XXXXXXX XXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXX
                    BBBB BBBBBBBBBBBBBBBBBBBBBBBBB BB BBBBBBBBBBBB
                        
                    BBBBB


                BBBBB
            XXXXXX
            XXXX XXXXXXXXXXXXXXXXXXXXX
                BBBBBBB BBBBBBBBBBBBBBBBBBBBBBBBBBBB BBBB BBBBBBBBBBBBBBBBB
            XXXXXX
        XXXXXX
    XXXXXX


BBBBBBBB
BBBBB BBBBBBBBB
    XXXXXXX XXXXXBBBBBB BBBBBBBBBBBBBBBBBBBBBBBBBXXXXXXXXXXX
    XXXXXXX XXXXXBBBBBB BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBXXXXXXXXXXX
    XXXXXXX XXXXXBBBBBB BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBXXXXXXXXXXX


    XXXXXXXX
        XXXXXXXXXXXXXXXXXXXXXXXXXX XX X
            XXX XXXX X FFFFFFFX

            XXX XXXXXXXXXXXXXX X XXXXXXXXXXXXXXXXX

            XXXXXXXX XXXXXXXXXXXXXXX X
                XXXXXXXX XXXXXXXXXXXXXXXXXX X
                    XXX XXXXXXXXXXX X XXX XXXXXXX
                    XXX XXXXXXXXX X XXX XXXXXXX
                    XXX XXXXXXXXX X XXXXXXXXXXXX
                    XXX XXXXXXXXXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                    XXX XXXXXXXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                    XX XXXXXXXXXXXXXXXXX XX XXXXXXXXXXXXXXX X
                        XXXXXXXXXXX X XXX XXXXXXXXXXXXXXXXXXXXXXX
                        XXXXXXXXX X XXX XXXXXXXXXXXXXXXXXXXXX
                    X XXXX X
                        XXXXXXXXXXXXXXXXXXXXXXXX
                        XXXXXXXXXXXXXXXXXXXXXXXXXX
                        XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX X XXX
                        XXXXXXXXXXXXXXXXXXXXXX
                        XXXXXXXXXXXXXXXXXXXXXXXX
                    X

                    XXXXXX X

                        XXXXXXXXXXXXXXXX XXXXXX
                        XXXXXXXXXXX XXXXX
                        XXXXXXXXXXXXXXXXX XXXXX
                        XXXXXXXXXXXXXXXXXXXX XXX
                        XXXXXXXXXX XXXXX
                        XXXXXXXXXX XXXXXXXXXXXX
                        XXXXXXXX XXXXXXXXXX
                        XXXXXXXXXXXXXXXXXXXX XXXXX

                        XXXXXXX X
                            X gettext(u'Today') XX XXXXXXXXXXXXX XXXXXXXXXXX
                            X gettext(u'Yesterday') XX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXX
                            X gettext(u'Last 7 Days') XX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXX XXXXXXXXXXXXXXXXXXX
                            X gettext(u'This Month') XX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                            X gettext(u'Last Month') XX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXX
                        XX

                        XXXXXX XXXXXXXX
                        XXXXXXXXXXXXXX XXXXX
                        XXXXXXXXXXXXXXXX XXXXXX
                        XXXXXXXXXXXXXX XXXXXXXX
                        XXXXXXXXXXX XXXXXXXXXX XXXXXXXX XXXXXXXXXXX
                        XXXXXXXXXXXX XXXXXXXXXX XXXXXXXXXXX XXXXXXXXXXX
                        XXXXXXX X
                            XXXXXXXXXX BB BBBBBBBBBBBBB XXXXX BBBBXXXXXBBBBBX
                            XXXXXXX XXXXXXXXXXXXX
                            XXXXXXXXXXX X gettext(u'Apply') XX
                            XXXXXXXXXX X gettext(u'From') XX
                            XXXXXXXX X gettext(u'To') XX
                            XXXXXXXXXXX X gettext(u'start date') XXX
                            XXXXXXXXX X gettext(u'end date') XXX
                            XXXXXXXXXXXX X gettext(u'Cancel') XX
                            XXXXXXXXXXXXXXXXX X gettext(u'choose') XX
                            XXXXXXXXX X
                        X
                    XX
                X

                XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXX XXXX XXXXXXX X
                    XXX XXXXXXXXX X XXXXXXXXXXXXXXX
                    XXX XXXXXXXXXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                    XXX XXXXXXXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                    XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX X XXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXX X XXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXX X X X X XXXXXXXXXXXXXXXXXXXXXXXXXXXXX X XXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                    XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX X XXX X XXXXXXXXXXXXXXXXXXXXXXXXXXX X XXX X XXXXXXXXXXXXXXXXXXXXXXXXXXX X X X X XXXXXXXXXXXXXXXXXXXXXXXXXXX X XXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXX
                    XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                XX

            X

            XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXX XXX XXXXXXXX XXX X
                XXXXXXXXXXXXXXXXXXX
                XXX XXXXX X XXXXXXXX
                XXX XXXXXXXXXXXXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

                XX XXXXXXXXXXXXXX XXX XXXXXXXXXXX X
                    XXXXXXXXXXXXXXXXXXXXXXXXXXX
                X XXXX XX XXXXXXXXXXXXXX XXX XXXXXXXXXXXXX X
                    XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                X

                XXX XX X XXXXXXXXXXXXXXXXXXXX
                XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXX XXX XXXXXX XXXXXXX
                XXXXXXXXXXXXXXXXXXXXXXXXXXX X XXXXXXXXXXXXXXXXXXXXXXXXXXX X XXXXXXXXXXXXXXXXXXXXXXXX XX XXX XXX XXX XXXXXX XXXXX


            XXX

            XXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXX X
                XXXXXXXXXXXXXXXXXXXXXXX
                XXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXX

                XXX XXXXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                XXX XXXXXXXXXXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                XXXXXXXXXXXXXXXXXXXXXXXXXXXXX

            X

            XXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXX X
                XXXXXXXXXXXXXXXXXXXXXXX
                XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX  XXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXX
                XXX XXXXXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
                XXX XXXXXXXXXXX X XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXX
                XXXXXXX X XXX XXXXXXXXXXXXX XXXXXXXXXXXXX

            X

            XXXXXXXXXXXXXXXXXXXXXXX
            XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
            XXXXXXXXXXXXXXXX


        XX
    XXXXXXXXX

BBBBBBBB



from collections import OrderedDict

from django import forms
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from . import app_settings
from .helpers import get_foreign_keys


class BaseReportForm:
    '''
    Holds basic function
    '''

    @property
    def media(self):
        from .app_settings import SLICK_REPORTING_FORM_MEDIA
        return forms.Media(css=SLICK_REPORTING_FORM_MEDIA.get('css', {}), js=SLICK_REPORTING_FORM_MEDIA.get('js', []))

    def get_filters(self):
        """
        Get the foreign key filters for report queryset, excluding crosstab ids, handled by `get_crosstab_ids()`
        :return: a dicttionary of filters to be used with QuerySet.filter(**returned_value)
        """
        _values = {}
        if self.is_valid():
            for key, field in self.foreign_keys.items():
                if key in self.cleaned_data and not key == self.crosstab_key_name:
                    val = self.cleaned_data[key]
                    if val:
                        val = [x for x in val.values_list('pk', flat=True)]
                        _values['%s__in' % key] = val
            return None, _values

    @cached_property
    def crosstab_key_name(self):
        """
        return the actual foreignkey field name by simply adding an '_id' at the end.
        This is hook is to customize this naieve approach.
        :return: key: a string that should be in self.cleaned_data
        """
        return f'{self.crosstab_model}_id'

    def get_crosstab_ids(self):
        """
        Get the crosstab ids so they can be sent to the report generator.
        :return: 
        """
        if self.crosstab_model:
            qs = self.cleaned_data.get(self.crosstab_key_name)
            return [x for x in qs.values_list('pk', flat=True)]
        return []

    def get_crosstab_compute_reminder(self):
        return self.cleaned_data.get('crosstab_compute_reminder', True)

    def get_crispy_helper(self, foreign_keys_map=None, crosstab_model=None, **kwargs):
        from crispy_forms.helper import FormHelper
        from crispy_forms.layout import Column, Layout, Div, Row, Field

        helper = FormHelper()
        helper.form_class = 'form-horizontal'
        helper.label_class = 'col-sm-2 col-md-2 col-lg-2'
        helper.field_class = 'col-sm-10 col-md-10 col-lg-10'
        helper.form_tag = False
        helper.disable_csrf = True
        helper.render_unmentioned_fields = True

        foreign_keys_map = foreign_keys_map or self.foreign_keys

        helper.layout = Layout(
            Row(
                Column(
                    Field('start_date'), css_class='col-sm-6'),
                Column(
                    Field('end_date'), css_class='col-sm-6'),
                css_class='raReportDateRange'),
            Div(css_class="mt-20", style='margin-top:20px')
        )

        # first add the crosstab model and its display reimder then the rest of the fields
        if self.crosstab_model:
            helper.layout.fields[1].append(Field(self.crosstab_key_name))
            if self.crosstab_display_compute_reminder:
                helper.layout.fields[1].append(Field('crosstab_compute_reminder'))

        for k in foreign_keys_map:
            if k != self.crosstab_key_name:
                helper.layout.fields[1].append(Field(k))

        return helper


def _default_foreign_key_widget(f_field):
    return {'form_class': forms.ModelMultipleChoiceField, 'required': False, }


def report_form_factory(model, crosstab_model=None, display_compute_reminder=True, fkeys_filter_func=None,
                        foreign_key_widget_func=None, excluded_fields=None):
    """
    Create a Report Form based on the report_model passed by
    1. adding a start_date and end_date fields
    2. extract all ForeignKeys on the report_model
    
    :param model: the report_model
    :param crosstab_model: crosstab model if any
    :param display_compute_reminder:  relevant only if crosstab_model is specified. Control if we show the check to
    display the rest.
    :param fkeys_filter_func:  a receives an OrderedDict of Foreign Keys names and their model field instances found on the model, return the OrderedDict that would be used
    :param foreign_key_widget_func: receives a Field class return the used widget like this {'form_class': forms.ModelMultipleChoiceField, 'required': False, }
    :param excluded_fields: a list of fields to be excluded from the report form
    :return:
    """
    foreign_key_widget_func = foreign_key_widget_func or _default_foreign_key_widget
    fkeys_filter_func = fkeys_filter_func or (lambda x: x)

    # gather foreign keys
    fkeys_map = get_foreign_keys(model)
    excluded_fields = excluded_fields or []
    for excluded in excluded_fields:
        del fkeys_map[excluded]

    fkeys_map = fkeys_filter_func(fkeys_map)

    fkeys_list = []
    fields = OrderedDict()

    fields['start_date'] = forms.DateTimeField(required=False, label=_('From date'),
                                               initial=app_settings.SLICK_REPORTING_DEFAULT_START_DATE,
                                               widget=forms.DateTimeInput(attrs={'autocomplete': "off"}))

    fields['end_date'] = forms.DateTimeField(required=False, label=_('To  date'),
                                             initial=app_settings.SLICK_REPORTING_DEFAULT_END_DATE,
                                             widget=forms.DateTimeInput(attrs={'autocomplete': "off"}))

    for name, f_field in fkeys_map.items():
        fkeys_list.append(name)

        fields[name] = f_field.formfield(**foreign_key_widget_func(f_field))

    if crosstab_model and display_compute_reminder:
        fields['crosstab_compute_reminder'] = forms.BooleanField(required=False,
                                                                 label=_('Display the crosstab reminder'),
                                                                 initial=True)

    new_form = type('ReportForm', (BaseReportForm, forms.BaseForm,),
                    {"base_fields": fields,
                     '_fkeys': fkeys_list,
                     'foreign_keys': fkeys_map,
                     'crosstab_model': crosstab_model,
                     'crosstab_display_compute_reminder': display_compute_reminder,
                     })
    return new_form



XXXXXXXXX XXXXX
BBBB BBBBBB
BBBB BBBBBBBBBBBBBBBBB

XXXXX XXXXXXXXXX
XXXXXX
    XXXX XXXXXXXX XXXX XXXX XXX
    XXXXX XXXXXXXXXXXXXXXX
    XXXXX XXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXX

    XXXX XXXXXXXXX XXX XXX
    XXXXX XXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
          XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXX

    XXXXX XXXXXXXXXXXXXXXX XXXXXXXXXXXXXXX
          XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
    XXXXXXXXXXXXX XXXXX XXXXXXXXXXXXXXXXX
XXXXXXX
XXXXXX
XXXX XXXXXXXXXXXXXXXXXX
    XXXX XXXXXXXXXXX XX
        BBBBB BBBBBBB

        BBBBBBBB
    XXXXXX
XXXXXX
XXXX XXXXXXXX XXXXXXXXXX XXX
XXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX


XXXXXXX XXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXXXX XXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
XXXXX XXXXXXXXXXXXXXXX XXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

XXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XXXXXXXXXXXXXXXXXX
XXXXXXX XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
        XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX XX X
    XXXXXXXXXXXXXXXXXXXXXX
XXXXXXXXXXX
BBBBB BBBBBBBBB
BBBBBBBB
XXXXXXX
XXXXXXX